#include "ai.h"
#include <ai_nodes.h>
#include <string.h>
#include <ldpk/ldpk.h>
#include <fstream>

// #include <iostream>
// #include <stdio.h>

using namespace std;
using namespace ldpk;

enum tdeParams {
                p_distort,
                p_model,
                p_distFile,
                p_focal_length,
                p_focus_distance,
                p_filmBack_W,
                p_filmBack_H,
                p_xOffset,
                p_yOffset,
                p_pixel_aspect,
                p_distortion,
                p_anamorphicSqueeze,
                p_xCurvature,
                p_yCurvature,
                p_quarticDist,
                p_dd2, p_ud2, p_vd2, p_qd4, p_ud4, p_vd4, p_d6, p_d8, p_phi, p_B,
                p_cx02d2, p_cy02d2, p_cx22d2, p_cy22d2, p_cx04d4, p_cy04d4, p_cx24d4, p_cy24d4,
                p_cx44d4, p_cy44d4, p_rot, p_sqx, p_sqy,
                p_cx06d6, p_cy06d6, p_cx26d6, p_cy26d6, p_cx46d6, p_cy46d6, p_cx66d6, p_cy66d6,
};

enum distModel
{
    CLASSIC = 0,
    RADIAL,
    FISHEYE,
    ANAMORPHIC,
    SQUEEZEROT
};

const char* model_enum[] =
{
    "3DE Classic LD Model",
    "3DE4 Radial - Standard, Degree 4",
    "3DE4 Anamorphic - Standard, Degree 4",
    "3DE4 Radial - Fisheye, Degree 8",
    "3DE4 Anamorphic, Degree 6",
    NULL
};

AI_CAMERA_NODE_EXPORT_METHODS(tdeLens)

node_parameters
{
    AiParameterBool("distort", true);
    AiParameterENUM("model", 0, model_enum);
    AiParameterSTR("distortion_file", "");
    AiParameterFlt("focal_length", 32.0f);
    AiParameterFlt("focus_distance", 100.0f);
    AiParameterFlt("filmBack_W", 2.6729f);
    AiParameterFlt("filmBack_H", 1.3365f);
    AiParameterFlt("xOffset", 0.0f);             // filmback x offset
    AiParameterFlt("yOffset", 0.0f);             // filmback y offset
    AiParameterFlt("pixel_aspect", 1.0f);
    AiParameterFlt("distortion", 0.0f);	         // Lens Distortion
    AiParameterFlt("anamorphicSqueeze", 1.0f);   // Anamorphic Squeeze
    AiParameterFlt("xCurvature", 0.0f);		     // X Curvature
    AiParameterFlt("yCurvature", 0.0f);          // Y Curvature
    AiParameterFlt("quarticDist", 0.0f);         // Quartic Term    

    AiParameterFlt("distortion_Degree2", 0.0f);
    AiParameterFlt("u_Degree2", 0.0f);
    AiParameterFlt("v_Degree2", 0.0f);
    AiParameterFlt("quartic_Degree4", 0.0f);
    AiParameterFlt("u_Degree4", 0.0f);
    AiParameterFlt("v_Degree4", 0.0f);
    AiParameterFlt("Degree6", 0.0f);
    AiParameterFlt("Degree8", 0.0f);
    AiParameterFlt("Phi_cyl_dir", 0.0f);
    AiParameterFlt("B_cyl_bend", 0.0f);

    AiParameterFlt("cx02d2", 0.0f);
    AiParameterFlt("cy02d2", 0.0f);
    AiParameterFlt("cx22d2", 0.0f);
    AiParameterFlt("cy22d2", 0.0f);
    AiParameterFlt("cx04d4", 0.0f);
    AiParameterFlt("cy04d4", 0.0f);
    AiParameterFlt("cx24d4", 0.0f);
    AiParameterFlt("cy24d4", 0.0f);
    AiParameterFlt("cx44d4", 0.0f);
    AiParameterFlt("cy44d4", 0.0f);
    AiParameterFlt("lensRotation", 0.0f);
    AiParameterFlt("squeeze_x", 1.0f);
    AiParameterFlt("squeeze_y", 1.0f);

    AiParameterFlt("cx06d6", 0.0f);
    AiParameterFlt("cy06d6", 0.0f);
    AiParameterFlt("cx26d6", 0.0f);
    AiParameterFlt("cy26d6", 0.0f);
    AiParameterFlt("cx46d6", 0.0f);
    AiParameterFlt("cy46d6", 0.0f);
    AiParameterFlt("cx66d6", 0.0f);
    AiParameterFlt("cy66d6", 0.0f);
}

typedef struct {
    bool distOn;
    int model;
    float fov;
    float hfb;
    float wfb;
    float pixel_aspect;
    float f;
    float focus_distance;
    float dist;
    float anamorphic;
    float xCurve;
    float yCurve;
    float delta_xfb, delta_yfb, q;
    float pixelAspect;
    const char * distFile;
    float dd2, ud2, vd2, qd4, ud4, vd4, d6, d8, phi, B;
    float cx02d2, cy02d2, cx22d2, cy22d2, cx04d4, cy04d4, cx24d4, cy24d4, cx44d4, cy44d4, rot, sqx, sqy;
    float cx06d6, cy06d6, cx26d6, cy26d6, cx46d6, cy46d6, cx66d6, cy66d6;
    double maxX, maxY;
    // motion blur status
    // bool mb;
} tdeLensData;

tdeLensData* lens = (tdeLensData*) AiMalloc(sizeof(tdeLensData));
plugin_loader pl;
model_parser mp(false);

bool ld_model()
{
    string modelPath;
    string envLens = getenv("LENSMODEL_PATH");

    if (lens->distOn)
    {
        if (lens->model < 0 || lens->model > 4)
        {
            // Disable distortion if an invalid model number was given
            AiMsgWarning("The distortion model is invalid!! Regular perspective camera will be used.");
            return false;
        }

        if (lens->model == 0)
        {
            modelPath = envLens + "/tde4_ldp_classic_3de_mixed" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                AiMsgWarning("tde4_ldp_classic_3de_mixed.so is not exist! Make sure you set env LENSMODEL_PATH");
                return false;
            }
        }
        if (lens->model == 1)
        {
            modelPath = envLens + "/tde4_ldp_radial_decentered_deg_4_cylindric" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                AiMsgWarning("tde4_ldp_radial_decentered_deg_4_cylindric.so is not exist! Make sure you set env LENSMODEL_PATH");
                return false;
            }
        }
        if (lens->model == 2)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                AiMsgWarning("tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy.so is not exist! Make sure you set env LENSMODEL_PATH");
                return false;
            }
        }
        if (lens->model == 3)
        {
            modelPath = envLens + "/tde4_ldp_radial_deg_8" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                AiMsgWarning("tde4_ldp_radial_deg_8.so is not exist! Make sure you set env LENSMODEL_PATH");
                return false;
            }
        }
        if (lens->model == 4)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_6" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                AiMsgWarning("tde4_ldp_anamorphic_deg_6.so is not exist! Make sure you set env LENSMODEL_PATH");
                return false;
            }
        }
    }
    pl.open_plugin(modelPath);

    // Set the parameters that are common for any lens
    pl.get_model()->setParameterValue("tde4_focal_length_cm", lens->f);
    pl.get_model()->setParameterValue("tde4_filmback_width_cm", lens->wfb);
    pl.get_model()->setParameterValue("tde4_filmback_height_cm", lens->hfb);
    pl.get_model()->setParameterValue("tde4_pixel_aspect", lens->pixel_aspect);
    pl.get_model()->setParameterValue("tde4_lens_center_offset_x_cm", lens->delta_xfb);
    pl.get_model()->setParameterValue("tde4_lens_center_offset_y_cm", lens->delta_yfb);
    pl.get_model()->setParameterValue("tde4_custom_focus_distance_cm", lens->focus_distance);

    switch (lens->model)
    {
    case 0:
        // Set the paramters that are specific to the 'Classic 3DE distortion' model
        pl.get_model()->setParameterValue("Distortion" , lens->dist);
        pl.get_model()->setParameterValue("Anamorphic Squeeze", lens->anamorphic);
        pl.get_model()->setParameterValue("Curvature X", lens->xCurve);
        pl.get_model()->setParameterValue("Curvature Y", lens->yCurve);
        pl.get_model()->setParameterValue("Quartic Distortion", lens->q);
        break;
    case 1:
        // Set the paramters that are specific to the 'Radial - Standard, Degree 4' model
        pl.get_model()->setParameterValue("Distortion - Degree 2" , lens->dd2);
        pl.get_model()->setParameterValue("U - Degree 2" , lens->ud2);
        pl.get_model()->setParameterValue("V - Degree 2" , lens->vd2);
        pl.get_model()->setParameterValue("Quartic Distortion - Degree 4" , lens->qd4);
        pl.get_model()->setParameterValue("U - Degree 4" , lens->ud4);
        pl.get_model()->setParameterValue("V - Degree 4" , lens->vd4);
        pl.get_model()->setParameterValue("Phi - Cylindric Direction" , lens->phi);
        pl.get_model()->setParameterValue("B - Cylindric Bending" , lens->B);
        break;
    case 2:
    	// Set the paramters that are specific to the 'Anamorphic - Standard, Degree 4' model
    	pl.get_model()->setParameterValue("Cx02 - Degree 2", lens->cx02d2);
        pl.get_model()->setParameterValue("Cy02 - Degree 2", lens->cy02d2);
        pl.get_model()->setParameterValue("Cx22 - Degree 2", lens->cx22d2);
        pl.get_model()->setParameterValue("Cy22 - Degree 2", lens->cy22d2);
        pl.get_model()->setParameterValue("Cx04 - Degree 4", lens->cx04d4);
        pl.get_model()->setParameterValue("Cy04 - Degree 4", lens->cy04d4);
        pl.get_model()->setParameterValue("Cx24 - Degree 4", lens->cx24d4);
        pl.get_model()->setParameterValue("Cy24 - Degree 4", lens->cy24d4);
        pl.get_model()->setParameterValue("Cx44 - Degree 4", lens->cx44d4);
        pl.get_model()->setParameterValue("Cy44 - Degree 4", lens->cy44d4);
        pl.get_model()->setParameterValue("Lens Rotation", lens->rot);
        pl.get_model()->setParameterValue("Squeeze-X", lens->sqx);
        pl.get_model()->setParameterValue("Squeeze-Y", lens->sqy);
	break;
    case 3:
        // Set the paramters that are specific to the 'Radial - Fisheye, Degree 8' model
        pl.get_model()->setParameterValue("Distortion - Degree 2", lens->dd2);
        pl.get_model()->setParameterValue("Quartic Distortion - Degree 4", lens->qd4);
        pl.get_model()->setParameterValue("Degree 6" , lens->d6);
        pl.get_model()->setParameterValue("Degree 8" , lens->d8);
        break;
    case 4:
        pl.get_model()->setParameterValue("Cx02 - Degree 2", lens->cx02d2);
        pl.get_model()->setParameterValue("Cy02 - Degree 2", lens->cy02d2);
        pl.get_model()->setParameterValue("Cx22 - Degree 2", lens->cx22d2);
        pl.get_model()->setParameterValue("Cy22 - Degree 2", lens->cy22d2);
        pl.get_model()->setParameterValue("Cx04 - Degree 4", lens->cx04d4);
        pl.get_model()->setParameterValue("Cy04 - Degree 4", lens->cy04d4);
        pl.get_model()->setParameterValue("Cx24 - Degree 4", lens->cx24d4);
        pl.get_model()->setParameterValue("Cy24 - Degree 4", lens->cy24d4);
        pl.get_model()->setParameterValue("Cx44 - Degree 4", lens->cx44d4);
        pl.get_model()->setParameterValue("Cy44 - Degree 4", lens->cy44d4);
        pl.get_model()->setParameterValue("Cx06 - Degree 6", lens->cx06d6);
        pl.get_model()->setParameterValue("Cy06 - Degree 6", lens->cy06d6);
        pl.get_model()->setParameterValue("Cx26 - Degree 6", lens->cx26d6);
        pl.get_model()->setParameterValue("Cy26 - Degree 6", lens->cy26d6);
        pl.get_model()->setParameterValue("Cx46 - Degree 6", lens->cx46d6);
        pl.get_model()->setParameterValue("Cy46 - Degree 6", lens->cy46d6);
        pl.get_model()->setParameterValue("Cx66 - Degree 6", lens->cx66d6);
        pl.get_model()->setParameterValue("Cy66 - Degree 6", lens->cy66d6);
        break;
    }
    pl.get_model()->initializeParameters();
    // Parsing Distortion Parameters from file if one exists,
    // This overrides any local parameters.
    // I found this useful in many cases and makes consistent workflow between Arnold and 3DE
    // Also the one could use it as a lens Specs file for other non-tracked scenes.
    const char * distFile = lens->distFile;
    string df = distFile;
    if (df.length() > 1) // null makes some troubles inside Maya if a file was entered and deleted
                         // here, a distortion filename length/path should be greater than one character.
    {
        ifstream dFile(distFile);
        if(dFile.fail())
        {
            AiMsgWarning("Cannot open the Distortion parameters file! Local parameters will be used instead.");
        }
        else
        {
            // We get the Common Lens parameters from local parameters if were not been set
            // in the param file.
            mp.set_model(pl.get_model());
            mp.parse(dFile);
            return true;
        }
    }
    return true;
}

void distort(float *xdn, float *ydn)
{
    double x_dn0, y_dn0;
    double x_dn = *xdn;
    double y_dn = *ydn;

    pl.get_model()->undistort(x_dn, y_dn, x_dn0, y_dn0);

    *xdn = x_dn0;
    *ydn = y_dn0;
}

node_initialize
{
    AiCameraInitialize(node, NULL);

    // Deny other cameras in the scene from being initialized or updated! And prevent the renCam param
    // from being overwritten by those inactive cameras.
    // I think I found a API bug here, I expected that only the active camera should be initialized
    // and updated!! No need to initialize and update all the non-active cameras in the scene.
    // Get the Active camera pointer
    AtNode *CamPtr = AiUniverseGetCamera();
    // Get the name of the Active Rendering Camera
    const char* renCam = AiNodeGetName(CamPtr);
    // Get the name of the currently evaluating camera
    const char* curCam = AiNodeGetName(node);
    if (curCam != renCam)
        return;
    //----- The previous protection code should be removed when the API bug is fixed -----//

    lens->f = params[p_focal_length].FLT / 10.0f;   // focal length in cm
    lens->wfb = params[p_filmBack_W].FLT * 2.54f;   // film back cm
    lens->hfb = params[p_filmBack_H].FLT * 2.54f;   // film width cm

    // Calculating the camera FOV
    // Fix Me: We should get the FOV array directly from the API when it's supported.
    double df = sqrt(lens->wfb*lens->wfb + lens->hfb*lens->hfb);
    lens->fov =2.0f * atan(lens->wfb / (2.0f * lens->f));
    lens->fov *= AI_RTOD;

    lens->delta_xfb = params[p_xOffset].FLT * 2.54f;    // film Ofsset X in cm
    lens->delta_yfb = params[p_yOffset].FLT * 2.54f;    // film Offset Y in cm
    lens->focus_distance = params[p_focus_distance].FLT;
    lens->pixel_aspect = params[p_pixel_aspect].FLT;

    lens->distOn = params[p_distort].BOOL;
    lens->model = params[p_model].INT;

    if (lens->distOn)
    {
        // Calculating the screen window Min/Max
        // Fix Me: We should get it using AiNodeGet... when the API add support for this.
        lens->maxX =  1.0f;
        lens->maxY = lens->hfb/lens->wfb;
        lens->distFile = params[p_distFile].STR;

        if (lens->model == 0)
        {
            lens->dist = params[p_distortion].FLT;
            lens->anamorphic = params[p_anamorphicSqueeze].FLT;
            lens->xCurve = params[p_xCurvature].FLT;
            lens->yCurve = params[p_yCurvature].FLT;
            lens->q = params[p_quarticDist].FLT;
        }
        if (lens->model == 1 || lens->model == 3)
        {
            lens->dd2 = params[p_dd2].FLT;
            lens->ud2 = params[p_ud2].FLT;
            lens->vd2 = params[p_vd2].FLT;
            lens->qd4 = params[p_qd4].FLT;
            lens->ud4 = params[p_ud4].FLT;
            lens->vd4 = params[p_vd4].FLT;
            lens->d6 = params[p_d6].FLT;
            lens->d8 = params[p_d8].FLT;
            lens->phi = params[p_phi].FLT;
            lens->B = params[p_B].FLT;
        }
        if (lens->model == 2)
        {
            lens->cx02d2 = params[p_cx02d2].FLT;
            lens->cy02d2 = params[p_cy02d2].FLT;
            lens->cx22d2 = params[p_cx22d2].FLT;
            lens->cy22d2 = params[p_cy22d2].FLT;
            lens->cx04d4 = params[p_cx04d4].FLT;
            lens->cy04d4 = params[p_cy04d4].FLT;
            lens->cx24d4 = params[p_cx24d4].FLT;
            lens->cy24d4 = params[p_cy24d4].FLT;
            lens->cx44d4 = params[p_cx44d4].FLT;
            lens->cy44d4 = params[p_cy44d4].FLT;
            lens->rot = params[p_rot].FLT;
            lens->sqx = params[p_sqx].FLT;
            lens->sqy = params[p_sqy].FLT;
        }
        if (lens->model == 4)
        {
            lens->cx02d2 = params[p_cx02d2].FLT;
            lens->cy02d2 = params[p_cy02d2].FLT;
            lens->cx22d2 = params[p_cx22d2].FLT;
            lens->cy22d2 = params[p_cy22d2].FLT;
            lens->cx04d4 = params[p_cx04d4].FLT;
            lens->cy04d4 = params[p_cy04d4].FLT;
            lens->cx24d4 = params[p_cx24d4].FLT;
            lens->cy24d4 = params[p_cy24d4].FLT;
            lens->cx44d4 = params[p_cx44d4].FLT;
            lens->cy44d4 = params[p_cy44d4].FLT;
            lens->cx06d6 = params[p_cx06d6].FLT;
            lens->cy06d6 = params[p_cy06d6].FLT;
            lens->cx26d6 = params[p_cx26d6].FLT;
            lens->cy26d6 = params[p_cy26d6].FLT;
            lens->cx46d6 = params[p_cx46d6].FLT;
            lens->cy46d6 = params[p_cy46d6].FLT;
            lens->cx66d6 = params[p_cx66d6].FLT;
            lens->cy66d6 = params[p_cy66d6].FLT;
        }

        // load the distortion model and initialize the distortion argument
        bool doIt = ld_model();
        if (!doIt)
            lens->distOn = false;
    }

}

node_update
{
    /*
    if (lens->mb)
    {
        AiCameraUpdate(node, true);
    } else {
        AiCameraUpdate(node, false);
    }
    */

    // FIXME: Don't update the node unless motion blur is enabled. This way, params will be
    //        parsed only once per render session if motion blur is disabled
    AiCameraUpdate(node, true);
}

camera_create_ray
{
    // Fix Me: We should read the FOV array from the API relative_time when the API support exporting
    // the FOV array, so, animated FOV would be respected by the Motion Blur.
    float tan_fov = tanf(lens->fov * AI_DTOR  / 2);
    AtPoint pdn;
    pdn.x = input->sx;
    pdn.y = input->sy;

    if (lens->distOn)
    {
        // normalize pdn from 0 to 1 instead of -min to +max
        pdn.x = (pdn.x + lens->maxX) / (2.0f *lens->maxX);
        pdn.y = (pdn.y + lens->maxY) / (2.0f *lens->maxY);
        // pdn.y = (input->sy + (lens->hfb/lens->wfb));

        distort(&pdn.x, &pdn.y);

        // inverse the normalization function to reformat pdn form -min to +max
        pdn.x = (pdn.x - 0.5f) * 2.0f * lens->maxX;
        pdn.y = (pdn.y - 0.5f) * 2.0f * lens->maxY;
    }
    pdn.x += (lens->delta_xfb);
    pdn.y += (lens->delta_yfb);

    AtPoint p = { pdn.x * tan_fov, pdn.y * tan_fov, 1 };
    output->dir = AiV3Normalize(p - output->origin);
    output->dir.z *= -1;


    // scale derivatives
    float dsx = input->dsx * tan_fov;
    float dsy = input->dsy * tan_fov;

    // Not sure if I need the following, but it's harmless to be impleted!
    AtVector d = p;
    double d_dot_d = AiV3Dot(d, d);
    double temp = 1.0 / sqrt(d_dot_d * d_dot_d * d_dot_d);

    output->dDdx.x = (d_dot_d * dsx - (d.x * dsx) * d.x) * temp;
    output->dDdx.y = (              - (d.x * dsx) * d.y) * temp;
    output->dDdx.z = (              - (d.x * dsx) * d.z) * temp;

    output->dDdy.x = (              - (d.y * dsy) * d.x) * temp;
    output->dDdy.y = (d_dot_d * dsy - (d.y * dsy) * d.y) * temp;
    output->dDdy.z = (              - (d.y * dsy) * d.z) * temp;
}

node_finish
{
   AiCameraDestroy(node);
}

node_loader
{
   if (i != 0)
      return false;

   node->methods     = tdeLens;
   node->output_type = AI_NODE_CAMERA;
   node->name        = "bs_tdeLens";
   node->node_type   = AI_NODE_CAMERA;
   strcpy(node->version, AI_VERSION);
   return true;
}
